#!/usr/bin/env python3
"""
This script is for the set-up of the controller "CTR-52" manufactured by MBJ Imaging.
The script uses the external modules "time" and "pyModbusTCP" to communicate with the device.
The defined testcases should give an impression on how to set up the CTR-52.
If you want to switch between modes turn off the channels before switching to another mode.

The function "complete_information" gives you all the necessary information of the controller.
"""

import struct
import time
from pyModbusTCP.client import ModbusClient

# define some constants for the test
ipaddr = '192.168.0.99'  # Standard address is 192.168.0.99
c = ModbusClient(host=ipaddr, port=502, auto_open=True, auto_close=True, debug=False)

# Coils general (Function code read 01 / write 05)
coil_restoresettings = 0
coil_flash_voltagemode = 2
coil_reseterrors = 12
# Coils channel specific (Function code read 01 / write 05)
coil_channel_sw_trigger = 20
coil_channel_tune_current = 21
coil_channel_invert_trigger = 22

# Discrete channel specific (Function code read 02)
discrete_channel_lightdetected = 20

# Holding general (Function code read 03 / write 06)
holding_ipadress = 0
holding_ipadress_low = 1
holding_subnetmask = 2
holding_storevalues = 4
holding_gaptime_low = 5
holding_gaptime_high = 6
holding_sequencemode = 8
# Holding channel specific (Function code read 03 / write 06)
holding_channel_mode = 20
holding_channel_triggerOR = 21
holding_channel_triggerAND = 22
holding_channel_current = 23
holding_channel_flashlength = 24
holding_channel_voltage_low = 26
holding_channel_voltage_high = 27

# INFORMATION - ONLY READ
# Input general (Function code read 04)
input_lightchannels = 0
input_HWverion = 1
input_FWversion = 2
input_IOprotocolversion = 3
input_productionweek = 4
input_errorcode = 5
input_mincurrent = 9
input_maxcurrent = 10
input_maxcurrentflash = 11
input_minflashtime = 12
input_maxflashtimelow = 13
input_maxflashtimehigh = 14
input_maxcapacity = 15
input_bootsequences = 16
# Input channel specific (Function code read 04)
input_channel_current = 20
input_channel_lastflashtime = 21
input_channel_voltage = 22
input_channel_tuningstatus = 23
input_channel_errorcode = 24

channeloffset = 10


def complete_information():
    print('------------ General Information about the controller --------------')
    hw_version = c.read_input_registers(input_HWverion)[0]
    print(f"HW version is: {divmod(hw_version, 256)[0]}.{divmod(hw_version, 256)[1]}")
    fw_version = c.read_input_registers(input_FWversion)[0]
    print(f"FW version is: {divmod(fw_version, 256)[0]}.{divmod(fw_version, 256)[1]}")
    lightchannels = c.read_input_registers(input_lightchannels)[0]
    print('Number of lightchannels is: ', lightchannels)
    prodweek = c.read_input_registers(input_productionweek, 1)[0]
    year, week = divmod(prodweek, 1 << 8)
    print('Production year is:', year)
    print('Production week is:', week)

    errorcode = c.read_input_registers(input_errorcode)[0]
    print('General Errorcode is: ', errorcode)
    if errorcode != 0:
        print('---------- Channel Errorcodes -----------')
        for i in range(4):
            errorcode_chan = c.read_input_registers(input_channel_errorcode + channeloffset * i, 1)[0]
            print('Errorcode Channel', i, ':', errorcode_chan)
    else:
        print('No channel errorcodes present!')

    mincurrent = c.read_input_registers(input_mincurrent)[0]
    print('Min. allowed current is:', mincurrent, 'mA')
    maxcurrent = c.read_input_registers(input_maxcurrent)[0]
    print('Max. allowed current is:', maxcurrent, 'mA')
    maxcurrentflash = c.read_input_registers(input_maxcurrentflash)[0]
    print('Max. allowed flash current is:', maxcurrentflash, 'mA')

    # Combination of two register values, one register is one word (16 Bit) long
    minflashtime = c.read_input_registers(input_minflashtime)[0]
    print('Min. allowed flashtime is:', minflashtime, 'us')
    maxflashtimelow = c.read_input_registers(input_maxflashtimelow, 1)[0]
    maxflashtimehigh = c.read_input_registers(input_maxflashtimehigh, 1)[0]
    # Convert to word, fill with zeros to get 16 Bit
    str_maxflashtimelow = format(maxflashtimelow, 'b').zfill(16)
    str_maxflashtimehigh = format(maxflashtimehigh, 'b').zfill(16)
    combined = [str_maxflashtimelow, str_maxflashtimehigh]
    str_maxflashtime = "".join(combined)
    print('Max. allowed flashtime is:', int(str_maxflashtime, 2) / 1000000, 's')

    gaptime = c.read_holding_registers(holding_gaptime_high, 1)[0]
    print('Gaptime is:', gaptime, 'us')
    maxcapacity = c.read_input_registers(input_maxcapacity)[0]
    print('Max. allowed capacity is: ', maxcapacity, 'mAs')
    bootsequences = c.read_input_registers(input_bootsequences)[0]
    print('Number bootsequences since first start: ', bootsequences)


def channel_off(channel):
    c.write_single_register(holding_channel_mode + channeloffset * channel, 0)
    time.sleep(0.2)


def channel_steady(channel):
    c.write_single_register(holding_channel_mode + channeloffset * channel, 1)
    time.sleep(0.2)


def channel_flash(channel):
    c.write_single_register(holding_channel_mode + channeloffset * channel, 2)
    time.sleep(0.2)


def channel_auto(channel):
    c.write_single_register(holding_channel_mode + channeloffset * channel, 3)
    time.sleep(0.2)


def channel_discharge(channel):
    c.write_single_register(holding_channel_mode + channeloffset * channel, 4)
    time.sleep(0.2)


def channel_read_mode(channel):
    currentmode = c.read_holding_registers(holding_channel_mode + channeloffset * channel, 1)
    return currentmode


def channel_tune(channel):
    c.write_single_coil(coil_channel_tune_current + channeloffset * channel, True)
    print('Tuning channel', channel)
    time.sleep(2)


def channel_set_current(channel, current):
    c.write_single_register(holding_channel_current + channeloffset * channel, current)
    print('Current Channel', channel, 'set to:', current)
    time.sleep(3)


def channel_read_current(channel):
    current = c.read_holding_registers(holding_channel_current + channeloffset * channel, 1)
    return current


def channel_write_voltage(channel, voltage):
    c.write_single_register(holding_channel_voltage_high + channeloffset * channel, voltage)
    print('Write Channel', channel, 'voltage:', voltage, 'V')
    time.sleep(0.2)


def channel_read_voltage(channel):
    voltage = c.read_holding_registers(holding_channel_voltage_high + channeloffset * channel, 1)
    return voltage


def channel_write_pulselength(channel, pulselength):
    c.write_multiple_registers(holding_channel_flashlength + channeloffset * channel, [0, pulselength])
    print('Write Channel', channel, 'pulselength:', pulselength, 'us')
    time.sleep(0.2)


def channel_read_pulselength(channel):
    flashlength = c.read_holding_registers(holding_channel_flashlength + channeloffset * channel, 1)
    return flashlength


def channel_read_voltages(channel):
    voltagehigh = c.read_holding_registers(holding_channel_voltage_high + channeloffset * channel, 1)
    time.sleep(0.2)
    return voltagehigh


def channel_read_live_currents(channel):
    livecurrent = c.read_input_registers(input_channel_current + channeloffset * channel, 1)
    time.sleep(0.2)
    return livecurrent


def channel_read_live_voltages(channel):
    livevoltage = c.read_input_registers(input_channel_voltage + channeloffset * channel, 1)
    time.sleep(0.2)
    return livevoltage


def read_voltagemode():
    voltagemode = c.read_coils(coil_flash_voltagemode, 1)[0]
    # Can be True or False
    return voltagemode


def activate_voltagemode():
    c.write_single_coil(coil_flash_voltagemode, True)
    voltagemode = c.read_coils(coil_flash_voltagemode, 1)[0]
    print('Voltagemode flash mode activated.', int(voltagemode))
    time.sleep(0.2)


def deactivate_voltagemode():
    c.write_single_coil(coil_flash_voltagemode, False)
    voltagemode = c.read_coils(coil_flash_voltagemode, 1)[0]
    print('Voltagemode flash mode deactivated.', int(voltagemode))
    time.sleep(0.2)


def write_sequence_mode(sequence):
    if 1 <= sequence <= 3:
        c.write_single_register(holding_sequencemode, sequence)
        print('Sequencemode', sequence, ': set')
        print('Only trigger channel 1 active. All channels in flash mode.')
    elif sequence == 0:
        c.write_single_register(holding_sequencemode, sequence)
        print('Sequence mode deactivated.')
    else:
        print('Input invalid.')
    time.sleep(0.2)


def read_sequence_mode():
    current_sequence = c.read_holding_registers(holding_sequencemode, 1)
    return current_sequence


def write_gaptime(timeinus):
    c.write_single_register(holding_gaptime_high, timeinus)
    # print('Write gaptime', timeinus)
    time.sleep(0.2)


def read_gaptime():
    gaptime = c.read_holding_registers(holding_gaptime_high, 1)
    return gaptime


def mask_channel_to_trg(channel, triggermask):
    c.write_single_register(holding_channel_triggerOR + channeloffset * channel, triggermask)
    print('Masking done. Channel', channel, 'to bitmask', triggermask)
    time.sleep(0.2)


def reset_errors():
    c.write_single_coil(coil_reseterrors, True)
    print('Reset errorcodes successful.')
    time.sleep(1)


def store_all_values():
    c.write_single_register(holding_storevalues, 12345)
    print('Values stored.')
    time.sleep(1)


def restore_settings():
    c.write_single_coil(coil_restoresettings, True)
    time.sleep(1)


def channel_change_mode(channel, mode):
    if (read_voltagemode() and mode == 1) or (read_voltagemode() and mode == 3):
        mode = 5

    match mode:
        case 0:
            print('Channel', channel, 'Mode: OFF')
            channel_off(channel)

        case 1:
            print('Channel', channel, 'Mode: Steady')
            channel_steady(channel)

        case 2:
            print('Channel', channel, 'Mode: Flash')
            channel_flash(channel)

        case 3:
            print('Channel', channel, 'Mode: Auto')
            channel_auto(channel)

        case 4:
            print('Channel', channel, 'Mode: Discharge')

        case 5:
            print('Voltagemode has to be deactivated first.')


def channel_change_current(channel, current):
    if read_voltagemode():
        print('To change current deactivate Voltagemode.')

    else:
        channel_set_current(channel, current)


def change_ip_address():
    byte1 = 192
    byte2 = 168
    byte3 = 0
    byte4 = 99

    tobytes = bytearray([byte1, byte2, byte3, byte4])  # Replace byte1, byte2, byte3, byte4 with your actual byte values

    # little endian
    value1 = struct.unpack('<H', tobytes[2:4])[0]  # Unpack bytes 2 and 3 as an unsigned short (ushort)
    value2 = struct.unpack('<H', tobytes[0:2])[0]  # Unpack bytes 0 and 1 as an unsigned short (ushort)

    print(value1)  # Result of the first computation
    print(value2)  # Result of the second computation

    combined_values = [value1, value2]
    print(combined_values)

    c.write_multiple_registers(holding_ipadress, combined_values)


def testcase_sequence():
    deactivate_voltagemode()

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    write_gaptime(10)

    channel_write_pulselength(0, 300)
    channel_write_pulselength(1, 300)
    channel_write_pulselength(2, 300)
    channel_write_pulselength(3, 300)

    write_sequence_mode(2)

    channel_change_mode(0, 2)
    channel_change_mode(1, 2)
    channel_change_mode(2, 2)
    channel_change_mode(3, 2)

    channel_change_current(0, 1000)
    channel_change_current(1, 1000)
    channel_change_current(2, 1000)
    channel_change_current(3, 1000)

    time.sleep(5)

    write_sequence_mode(1)

    channel_change_mode(0, 2)
    channel_change_mode(1, 2)
    channel_change_mode(2, 2)
    channel_change_mode(3, 2)

    channel_change_current(0, 1200)
    channel_change_current(1, 500)
    channel_change_current(2, 1200)
    channel_change_current(3, 500)

    time.sleep(5)

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)


def testcase_steady():
    deactivate_voltagemode()

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    channel_change_mode(0, 1)
    channel_change_mode(1, 1)
    channel_change_mode(2, 1)
    channel_change_mode(3, 1)

    time.sleep(2)
    test = channel_read_live_currents(0)
    print("Channel 1", test)
    test2 = channel_read_live_currents(1)
    print("Channel 2", test2)
    channel_read_live_currents(2)
    channel_read_live_currents(3)

    time.sleep(5)

    channel_change_current(0, 100)
    channel_change_current(1, 100)
    channel_change_current(2, 100)
    channel_change_current(3, 100)

    time.sleep(2)
    channel_read_live_currents(0)
    channel_read_live_currents(1)
    channel_read_live_currents(2)
    channel_read_live_currents(3)

    time.sleep(4)

    channel_change_current(0, 70)
    channel_change_current(1, 90)
    channel_change_current(2, 60)
    channel_change_current(3, 50)

    time.sleep(2)
    channel_read_live_currents(0)
    channel_read_live_currents(1)
    channel_read_live_currents(2)
    channel_read_live_currents(3)

    time.sleep(4)

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    # Schnelles AN/AUS von LED
    channel_change_mode(0, 1)
    channel_change_mode(1, 1)
    channel_change_mode(2, 1)
    channel_change_mode(3, 1)

    time.sleep(4)

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)


def testcase_voltagemode_flash():
    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    deactivate_voltagemode()
    activate_voltagemode()
    write_sequence_mode(0)

    channel_read_voltages(0)
    channel_read_voltages(1)
    channel_read_voltages(2)
    channel_read_voltages(3)

    write_gaptime(10)

    mask_channel_to_trg(0, 1)
    mask_channel_to_trg(1, 2)
    mask_channel_to_trg(2, 2)
    mask_channel_to_trg(3, 2)

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    channel_write_pulselength(0, 300)
    channel_write_pulselength(1, 300)
    channel_write_pulselength(2, 300)
    channel_write_pulselength(3, 300)

    channel_write_voltage(0, 18000)
    channel_write_voltage(1, 18000)
    channel_write_voltage(2, 18000)
    channel_write_voltage(3, 18000)

    channel_change_mode(0, 2)
    channel_change_mode(1, 2)
    channel_change_mode(2, 2)
    channel_change_mode(3, 2)

    channel_read_voltages(0)
    channel_read_voltages(1)
    channel_read_voltages(2)
    channel_read_voltages(3)

    channel_read_live_voltages(0)
    channel_read_live_voltages(1)
    channel_read_live_voltages(2)
    channel_read_live_voltages(3)

    channel_read_live_currents(0)
    channel_read_live_currents(1)
    channel_read_live_currents(2)
    channel_read_live_currents(3)


def testcase_flashmode():
    deactivate_voltagemode()
    write_sequence_mode(0)

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    channel_change_mode(0, 2)
    channel_change_mode(1, 2)
    channel_change_mode(2, 2)
    channel_change_mode(3, 2)

    channel_write_pulselength(0, 30)
    channel_write_pulselength(1, 30)
    channel_write_pulselength(2, 30)
    channel_write_pulselength(3, 30)

    write_gaptime(3)

    mask_channel_to_trg(0, 0)
    mask_channel_to_trg(1, 1)
    mask_channel_to_trg(2, 0)
    mask_channel_to_trg(3, 1)

    time.sleep(2)
    channel_read_live_currents(0)
    channel_read_live_currents(1)
    channel_read_live_currents(2)
    channel_read_live_currents(3)

    channel_read_live_voltages(0)
    channel_read_live_voltages(1)
    channel_read_live_voltages(2)
    channel_read_live_voltages(3)

    time.sleep(5)

    channel_change_current(0, 1000)
    channel_change_current(1, 1000)
    channel_change_current(2, 1000)
    channel_change_current(3, 1000)

    channel_change_mode(0, 0)
    channel_change_mode(1, 0)
    channel_change_mode(2, 0)
    channel_change_mode(3, 0)

    # Schnelles AN/AUS von LED
    channel_change_mode(0, 2)
    channel_change_mode(1, 2)
    channel_change_mode(2, 2)
    channel_change_mode(3, 2)

    time.sleep(4)


if __name__ == '__main__':
    # testcase_flashmode()
    complete_information()
    # change_ip_address()
    # store_all_values()
    # restore_settings()
    testcase_steady()
    # reset_errors()
    # testcase_steady()
    # testcase_voltagemode_flash()
    # testcase_sequence()
